/*
 Copyright (c) Information Equipment co.,LTD.
 All right reserved.
 
 Code by JaeHyuk Cho <mailto:minzkn@infoeq.com> Made in KOREA

 http://minzkn.pe.ky

 See also : /usr/include/bits/ioctls.h
*/

#ifndef DEF_SOURCE_networkinfo_c 
#define DEF_SOURCE_networkinfo_c "networkinfo.c"

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "mz.h"

static t_MZ_NetworkRoute *__MZ_DestroyNetworkRoute__(t_MZ_NetworkRoute *s_NetworkRoute);
static t_MZ_NetworkInterface *__MZ_DestroyNetworkInterface__(t_MZ_NetworkInterface *s_NetworkInterface);
static t_MZ_NetworkRoute *__MZ_CreateNetworkRoute__(void);
static t_MZ_NetworkInterface *__MZ_CreateNetworkInterface__(void);
static t_MZ_NetworkInfo *__MZ_CreateNetworkInfo__(void);

t_MZ_NetworkInfo *MZ_GetNetworkInfo(t_MZ_NetworkInfo *s_NetworkInfo);
t_MZ_NetworkInfo *MZ_DestroyNetworkInfo(t_MZ_NetworkInfo *s_NetworkInfo);

/* [ Static function ] ==================================================================== */
static t_MZ_NetworkRoute *__MZ_DestroyNetworkRoute__(t_MZ_NetworkRoute *s_NetworkRoute)
{
 t_MZ_NetworkRoute *s_Back;
 while(s_NetworkRoute)
 {
  s_Back = s_NetworkRoute;
  s_NetworkRoute = s_NetworkRoute->Next;
  if(s_Back->Name)MZ_Free(s_Back->Name);
  if(s_Back->Destination)MZ_Free(s_Back->Destination);
  if(s_Back->Gateway)MZ_Free(s_Back->Gateway);
  if(s_Back->Mask)MZ_Free(s_Back->Mask);
  MZ_Free(s_Back);
 }
 return(s_NetworkRoute);
}

static t_MZ_NetworkInterface *__MZ_DestroyNetworkInterface__(t_MZ_NetworkInterface *s_NetworkInterface)
{
 t_MZ_NetworkInterface *s_Back;
 while(s_NetworkInterface)
 {
  s_Back = s_NetworkInterface;
  s_NetworkInterface = s_NetworkInterface->Next;
  if(s_Back->Name     )MZ_Free(s_Back->Name);
  if(s_Back->IP       )MZ_Free(s_Back->IP);
  if(s_Back->Broadcast)MZ_Free(s_Back->Broadcast);
  if(s_Back->Netmask  )MZ_Free(s_Back->Netmask);
  if(s_Back->MAC      )MZ_Free(s_Back->MAC);
  MZ_Free(s_Back);
 }
 return(s_NetworkInterface);
}

static t_MZ_NetworkRoute *__MZ_CreateNetworkRoute__(void)
{
 t_MZ_NetworkRoute *s_Return;
 s_Return = (t_MZ_NetworkRoute *)MZ_Alloc(sizeof(t_MZ_NetworkRoute));
 if(s_Return)
 {
  s_Return->Next = (t_MZ_NetworkRoute *)0;
  s_Return->Name = s_Return->Destination = s_Return->Gateway = s_Return->Mask = (char *)0;
  s_Return->Flags = s_Return->RefCnt = s_Return->Use, s_Return->Metric = 0;
 }
 return(s_Return);
}

static t_MZ_NetworkInterface *__MZ_CreateNetworkInterface__(void)
{
 t_MZ_NetworkInterface *s_Return;
 s_Return = (t_MZ_NetworkInterface *)MZ_Alloc(sizeof(t_MZ_NetworkInterface));
 if(s_Return)
 {
  s_Return->Next = (t_MZ_NetworkInterface *)0;
  s_Return->Name = (char *)0;
  s_Return->IP = s_Return->Broadcast = s_Return->Netmask = s_Return->MAC = (char *)0;
  s_Return->Flags = s_Return->MTU = s_Return->Metric = 0;
 }
 return(s_Return);
}

static t_MZ_NetworkInfo *__MZ_CreateNetworkInfo__(void)
{
 t_MZ_NetworkInfo *s_Return;
 s_Return = (t_MZ_NetworkInfo *)MZ_Alloc(sizeof(t_MZ_NetworkInfo));
 if(s_Return)
 {
  s_Return->Route     = (t_MZ_NetworkRoute *)0;
  s_Return->Interface = (t_MZ_NetworkInterface *)0;
 }
 return(s_Return);
}

/* [ Public function ] ==================================================================== */
t_MZ_NetworkInfo *MZ_GetNetworkInfo(t_MZ_NetworkInfo *s_NetworkInfo)
{
 const char *c_Proc_NetworkRoute = "/proc/net/route";
 const char *c_Proc_NetworkDevice = "/proc/net/dev";
 int s_Handle, s_ReadBytes, s_Socket, s_Index, s_Byte;
 unsigned char s_Buffer[ 32 << 10 ];
 char *s_It;
 t_MZ_Parser *s_Parser = (t_MZ_Parser *)0;
 t_MZ_Node_Pointer *s_Line;
 t_MZ_NetworkRoute *s_TraceRoute, *s_NewRoute;
 t_MZ_NetworkInterface *s_TraceInterface, *s_NewInterface;
 unsigned long s_Destination, s_Gateway, s_Mask;
 int           s_Flags, s_RefCnt, s_Use, s_Metric;
 struct sockaddr_in s_SockAddress;
 /* Route parser */
 s_Handle = open(c_Proc_NetworkRoute, O_RDONLY);
 if(s_Handle >= 0)
 {
  do
  {
   s_ReadBytes = read(s_Handle, &s_Buffer[0], sizeof(s_Buffer));
   if(s_ReadBytes > 0)s_Parser = MZ_Parser_Line(s_Parser, &s_Buffer[0], s_ReadBytes);
  }while(s_ReadBytes == sizeof(s_Buffer));
  s_Parser = MZ_Parser_Line(s_Parser, (void *)0, 0); /* End of push */
  close(s_Handle);
  if(s_Parser)
  {
   if(s_NetworkInfo)s_NetworkInfo->Route     = __MZ_DestroyNetworkRoute__(s_NetworkInfo->Route);
   else s_NetworkInfo = __MZ_CreateNetworkInfo__();
   if(s_NetworkInfo)
   {
    s_Line = s_Parser->Line;
    s_TraceRoute = s_NetworkInfo->Route;
    while(s_Line)
    {
     if(s_Line->Pointer)
     {
      s_Index = 0;
      do
      {
       s_Byte = MZ_PeekByte(s_Line->Pointer, s_Index);
       if(s_Byte == '\0' || s_Byte == ' ' || s_Byte == '\t')break;
       s_Index++;
      }while(1);
      if(s_Index > 0)
      {
       if(MZ_PeekByte(s_Line->Pointer, s_Index) != '\0')MZ_PokeByte(s_Line->Pointer, s_Index++, '\0');
       do
       {
        s_Byte = MZ_PeekByte(s_Line->Pointer, s_Index);
        if(s_Byte == '\0' || (s_Byte != ' ' && s_Byte != '\t'))break;
        s_Index++;
       }while(1);
       if(sscanf(__MZ_PEEK__(s_Line->Pointer, s_Index), "%lx%lx%X%d%d%d%lx", &s_Destination, &s_Gateway, &s_Flags, &s_RefCnt, &s_Use, &s_Metric, &s_Mask) == 7)
       {
	s_NewRoute = __MZ_CreateNetworkRoute__();
        if(s_NewRoute)
	{
         s_NewRoute->Name = MZ_AllocString(s_Line->Pointer, (-1));
	 s_SockAddress.sin_addr.s_addr = s_Destination;
	 s_NewRoute->Destination = MZ_AllocString(inet_ntoa(s_SockAddress.sin_addr), (-1));
	 s_SockAddress.sin_addr.s_addr = s_Gateway;
	 s_NewRoute->Gateway = MZ_AllocString(inet_ntoa(s_SockAddress.sin_addr), (-1));
	 s_SockAddress.sin_addr.s_addr = s_Mask;
	 s_NewRoute->Mask = MZ_AllocString(inet_ntoa(s_SockAddress.sin_addr), (-1));
	 s_NewRoute->Flags = s_Flags;
	 s_NewRoute->RefCnt = s_RefCnt;
	 s_NewRoute->Use = s_Use;
	 s_NewRoute->Metric = s_Metric;
	 if(s_TraceRoute)
	 {
	  s_TraceRoute->Next = s_NewRoute;
	  s_TraceRoute = s_NewRoute;
	 }
	 else s_TraceRoute = s_NetworkInfo->Route = s_NewRoute;
	}	
       }
      }
     }
     s_Line = s_Line->Next;
    }
   }
   s_Parser = MZ_DestroyParser(s_Parser);
  } 
  else if(s_NetworkInfo)s_NetworkInfo = MZ_DestroyNetworkInfo(s_NetworkInfo);
  close(s_Handle);
 }
 else if(s_NetworkInfo)s_NetworkInfo = MZ_DestroyNetworkInfo(s_NetworkInfo);
 /* Network device parser */
 s_Handle = open(c_Proc_NetworkDevice, O_RDONLY);
 if(s_Handle >= 0)
 {
  do
  {
   s_ReadBytes = read(s_Handle, &s_Buffer[0], sizeof(s_Buffer));
   if(s_ReadBytes > 0)s_Parser = MZ_Parser_Line(s_Parser, &s_Buffer[0], s_ReadBytes);
  }while(s_ReadBytes == sizeof(s_Buffer));
  s_Parser = MZ_Parser_Line(s_Parser, (void *)0, 0); /* End of push */
  close(s_Handle);
  if(s_Parser)
  {
   if(s_NetworkInfo)s_NetworkInfo->Interface = __MZ_DestroyNetworkInterface__(s_NetworkInfo->Interface);
   else s_NetworkInfo = __MZ_CreateNetworkInfo__();
   if(s_NetworkInfo)
   {
    s_Line = s_Parser->Line;
    s_TraceInterface = s_NetworkInfo->Interface;
    while(s_Line)
    {
     if(s_Line->Pointer)
     {
      s_It = MZ_SearchString((char *)s_Line->Pointer, ":");
      if(s_It)
      {
       MZ_PokeByte(s_It++, 0, '\0');
       s_NewInterface = __MZ_CreateNetworkInterface__();
       if(s_NewInterface)
       {
	s_NewInterface->Name = MZ_StripString(s_Line->Pointer, " ", 1 | 2);
	if(s_NewInterface->Name)
	{
         if(s_TraceInterface)
	 {
          s_TraceInterface->Next = s_NewInterface;
          s_TraceInterface = s_NewInterface;
	 }
	 else s_TraceInterface = s_NetworkInfo->Interface = s_NewInterface;
         s_NewInterface = (t_MZ_NetworkInterface *)0; /* Release up */
	}
        if(s_NewInterface)s_NewInterface = __MZ_DestroyNetworkInterface__(s_NewInterface);
       }
      }
     }
     s_Line = s_Line->Next;
    }
   }
   s_Parser = MZ_DestroyParser(s_Parser);
  } 
  else if(s_NetworkInfo)s_NetworkInfo = MZ_DestroyNetworkInfo(s_NetworkInfo);
 }
 else if(s_NetworkInfo)s_NetworkInfo = MZ_DestroyNetworkInfo(s_NetworkInfo);
 /* Network info get parser */
 s_Socket = socket(AF_INET, SOCK_STREAM, 0);
 if(s_Socket >= 0)
 {
  struct ifconf s_ifconfig;
  struct ifreq *s_ifrequest;
  int s_RequestCount;
 
  s_RequestCount = 8; 
  MZ_ZeroMemory((void *)(&s_ifconfig), sizeof(s_ifconfig));
  do
  {
   if(s_ifconfig.ifc_buf)MZ_Free(s_ifconfig.ifc_buf);
   s_ifconfig.ifc_len = sizeof(struct ifreq) * s_RequestCount;
   s_ifconfig.ifc_buf = MZ_Alloc(s_ifconfig.ifc_len);
   if(s_ifconfig.ifc_buf)
   { 
    if(ioctl(s_Socket, SIOCGIFCONF, &s_ifconfig) == 0)
    {
     if(s_ifconfig.ifc_len < (sizeof(struct ifreq) * s_RequestCount))
     {
      if(s_ifconfig.ifc_req)
      {
       s_ifrequest = s_ifconfig.ifc_req;
       for(s_RequestCount = 0;s_RequestCount < s_ifconfig.ifc_len;s_RequestCount += sizeof(struct ifreq), s_ifrequest++)
       {
	if(s_NetworkInfo == (t_MZ_NetworkInfo *)0)s_NetworkInfo = __MZ_CreateNetworkInfo__();
        if(s_NetworkInfo)
	{
         if(s_NetworkInfo->Interface)
	 {
	  s_NewInterface = (t_MZ_NetworkInterface *)0;
          s_TraceInterface = s_NetworkInfo->Interface; 
	  while(s_TraceInterface)
	  {
           if(s_TraceInterface->Name)
	   {
            if(strcmp(s_TraceInterface->Name, s_ifrequest->ifr_name) == 0)
	    {
	     s_NewInterface = s_TraceInterface;
	     break; 
	    }
	   }
	   if(s_TraceInterface->Next)s_TraceInterface = s_TraceInterface->Next;
	   else
	   { /* New */
            s_NewInterface = s_TraceInterface->Next = __MZ_CreateNetworkInterface__();
	    if(s_NewInterface)s_NewInterface->Name = MZ_AllocString(s_ifrequest->ifr_name, (-1));
	    break;
	   }
	  } 
	 }
         else s_NewInterface = s_NetworkInfo->Interface = __MZ_CreateNetworkInterface__(); 

	 if(s_NewInterface)
	 {
	  if(s_NewInterface->IP)MZ_Free(s_NewInterface->IP);
	  if(ioctl(s_Socket, SIOCGIFADDR, s_ifrequest) == 0)
	  {
	   s_NewInterface->IP = MZ_AllocString(inet_ntoa(((struct sockaddr_in *)(&s_ifrequest->ifr_addr))->sin_addr), (-1));
	  }
	  else s_NewInterface->IP = (char *)0;
	  if(s_NewInterface->MAC)MZ_Free(s_NewInterface->MAC);
	  if(ioctl(s_Socket, SIOCGIFHWADDR, s_ifrequest) == 0)
	  {
	   s_NewInterface->MAC = MZ_SPrintf("%02X:%02X:%02X:%02X:%02X:%02X", 
			          (int)(s_ifrequest->ifr_hwaddr.sa_data[0] & 0xff),
			          (int)(s_ifrequest->ifr_hwaddr.sa_data[1] & 0xff),
			          (int)(s_ifrequest->ifr_hwaddr.sa_data[2] & 0xff),
			          (int)(s_ifrequest->ifr_hwaddr.sa_data[3] & 0xff),
			          (int)(s_ifrequest->ifr_hwaddr.sa_data[4] & 0xff),
			          (int)(s_ifrequest->ifr_hwaddr.sa_data[5] & 0xff)
			         );
	  }
	  else s_NewInterface->MAC = (char *)0;
	  if(s_NewInterface->Broadcast)MZ_Free(s_NewInterface->Broadcast);
	  if(ioctl(s_Socket, SIOCGIFBRDADDR, s_ifrequest) == 0)
	  {
	   s_NewInterface->Broadcast = MZ_AllocString(inet_ntoa(((struct sockaddr_in *)(&s_ifrequest->ifr_broadaddr))->sin_addr), (-1));
	  }
	  else s_NewInterface->Broadcast = (char *)0;
	  if(s_NewInterface->Netmask)MZ_Free(s_NewInterface->Netmask);
	  if(ioctl(s_Socket, SIOCGIFNETMASK, s_ifrequest) == 0)
	  {
	   s_NewInterface->Netmask = MZ_AllocString(inet_ntoa(((struct sockaddr_in *)(&s_ifrequest->ifr_netmask))->sin_addr), (-1));
	  }
	  else s_NewInterface->Netmask = (char *)0;
	  if(ioctl(s_Socket, SIOCGIFFLAGS, s_ifrequest) == 0)
	  {
#if 0
	  if(s_MyFlags & IFF_UP          )strcat(s_StringBuffer, "UP ");	
	  if(s_MyFlags & IFF_BROADCAST   )strcat(s_StringBuffer, "BROADCAST ");	
	  if(s_MyFlags & IFF_DEBUG       )strcat(s_StringBuffer, "DEBUG ");	
	  if(s_MyFlags & IFF_LOOPBACK    )strcat(s_StringBuffer, "LOOPBACK ");	
	  if(s_MyFlags & IFF_POINTOPOINT )strcat(s_StringBuffer, "POINTOPOINT ");	
	  if(s_MyFlags & IFF_NOTRAILERS  )strcat(s_StringBuffer, "NOTRAILERS ");	
	  if(s_MyFlags & IFF_RUNNING     )strcat(s_StringBuffer, "RUNNING ");	
	  if(s_MyFlags & IFF_NOARP       )strcat(s_StringBuffer, "NOARP ");	
	  if(s_MyFlags & IFF_PROMISC     )strcat(s_StringBuffer, "PROMISC ");	
	  if(s_MyFlags & IFF_ALLMULTI    )strcat(s_StringBuffer, "ALLMULTI ");	
	  if(s_MyFlags & IFF_SLAVE       )strcat(s_StringBuffer, "SLAVE ");	
	  if(s_MyFlags & IFF_MASTER      )strcat(s_StringBuffer, "MASTER ");	
	  if(s_MyFlags & IFF_MULTICAST   )strcat(s_StringBuffer, "MULTICAST ");	
#endif
	   s_NewInterface->Flags = s_ifrequest->ifr_flags;
	  }
	  else s_NewInterface->Flags = 0;
	  if(ioctl(s_Socket, SIOCGIFMTU, s_ifrequest) == 0) /* Get MTU size */
	  {
	   s_NewInterface->MTU = s_ifrequest->ifr_mtu;
	  }
	  else s_NewInterface->MTU = 0;
	  if(ioctl(s_Socket, SIOCGIFMETRIC, s_ifrequest) == 0)
	  {
	   s_NewInterface->Metric = s_ifrequest->ifr_metric ? s_ifrequest->ifr_metric : 1;
 	  }
	  else s_NewInterface->Metric = 0;
#if 0
	  if(ioctl(s_Socket, SIOCGIFDSTADDR, s_ifrequest) == 0) /* Get remote PA address */
	  {
	  }
#endif
#if 0
	  if(ioctl(s_Socket, SIOCGIFMEM, s_ifrequest) == 0) /* Get memory address (BSD) */
	  {
	  }
#endif
#if 0
	  if(ioctl(s_Socket, SIOCGIFENCAP, s_ifrequest) == 0) /* Get/Set encapsulations */
	  {
	  }
#endif
#if 0
	  if(ioctl(s_Socket, SIOCGIFSLAVE, s_ifrequest) == 0) /* Driver slaving support */
	  {
	  }
#endif
#if 0
	  if(ioctl(s_Socket, SIOCGIFINDEX, s_ifrequest) == 0) /* Name -> if_index mapping */
	  {
	  }
#endif
#if 0
	  if(ioctl(s_Socket, SIOCGIFPFLAGS, s_ifrequest) == 0)
	  {
	  }
#endif
#if 0
	  if(ioctl(s_Socket, SIOCGIFCOUNT, s_ifrequest) == 0) /* Get number of devices */
	  {
	  }
#endif
#if 0
	  if(ioctl(s_Socket, SIOCGIFBR, s_ifrequest) == 0) /* Bridging support */
	  {
	  }
#endif
#if 0
	  if(ioctl(s_Socket, SIOCGIFTXQLEN, s_ifrequest) == 0) /* Get the TX queue length */
	  {
	  }
#endif
#if 0
	  if(ioctl(s_Socket, SIOCGIFMAP, s_ifrequest) == 0) /* Get device parameters */
	  {
 	  }
#endif
	 }
	}
       } 
      }
      s_RequestCount = 0;
     }
     else s_RequestCount += 8;
    }
    else break;
   }
   else break;
  }while(s_RequestCount > 0 && s_RequestCount < 256);
  if(s_ifconfig.ifc_buf)MZ_Free(s_ifconfig.ifc_buf);
  close(s_Socket);
 }
 return(s_NetworkInfo);
}

t_MZ_NetworkInfo *MZ_DestroyNetworkInfo(t_MZ_NetworkInfo *s_NetworkInfo)
{
 if(s_NetworkInfo)
 {
  s_NetworkInfo->Route = __MZ_DestroyNetworkRoute__(s_NetworkInfo->Route);
  s_NetworkInfo->Interface = __MZ_DestroyNetworkInterface__(s_NetworkInfo->Interface);
  MZ_Free(s_NetworkInfo);
  s_NetworkInfo = (t_MZ_NetworkInfo *)0;
 }
 return(s_NetworkInfo);
}

#endif

/* End of source */
