/*
 Copyright (c) Information Equipment co.,LTD.
 All right reserved.
 
 Code by JaeHyuk Cho <mailto:minzkn@infoeq.com> Made in KOREA

 http://minzkn.pe.ky
*/

#ifndef DEF_SOURCE_string_c 
#define DEF_SOURCE_string_c "string.c"

#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include "mz.h"

int MZ_StringLength(const void *s_String);
void *MZ_AllocString(const void *s_String, int s_Size);
void *MZ_VSPrintf(const void *s_Format, va_list s_VaArg);
void *MZ_SPrintf(const void *s_Format, ...);
char *MZ_SearchString(const void *s_String, const void *s_This);
char *MZ_StripString(const void *s_String, const void *s_StripString, int s_Flag);

/* [ Static function ] ==================================================================== */

/* [ Public function ] ==================================================================== */
int MZ_StringLength(const void *s_String)
{
#ifdef WIN32
 if(s_String == (void *)0)return(0);
 return(strlen(s_String));
#else
 register int s_Return;
 if(s_String == (void *)0)return(0);
 __asm__ volatile(
  "\n\t"
  "cld\n\t"
  "repnz scasb %%es:(%1), %%al\n\t"
  "notl %2\n\t"
  "subl %3, %2\n\t"
  "decl %2\n\t"
  "\n\t"
  : "=a"(s_Return)
  : "D"(s_String), "a"(0), "c"(0xffffffff)
 );
 return(s_Return);
#endif
}

void *MZ_AllocString(const void *s_String, int s_Size)
{
 int s_Index;
 char *s_Return;
 if(s_String)
 {
  if(s_Size > 0)
  {
   s_Index = 0;
   while(s_Index < s_Size)
   {
    if(*((char *)__MZ_PEEK__(s_String, s_Index)) == '\0')break;
    s_Index++;
   }
  }
  else s_Index = MZ_StringLength(s_String);
  s_Return = (char *)MZ_Alloc(s_Index + 1);
  if(s_Return)
  {
   MZ_CopyMemory((void *)s_Return, s_String, s_Index);
   *((char *)__MZ_PEEK__(s_Return, s_Index)) = '\0';
  }
  return(s_Return);
 }
 return((void *)0);
}

void *MZ_VSPrintf(const void *s_Format, va_list s_VaArg)
{
#ifdef __STDC__
 void *s_Return;
 s_Return = MZ_Alloc( 1 << 10 );
 if(s_Return)vsprintf(s_Return, s_Format, s_VaArg);
 return(s_Return);
#else
 void *s_Return;
 int s_BufferSize = 256;
 do
 {
  s_Return = (void *)MZ_Alloc(s_BufferSize);
  if(s_Return)
  {
   if(vsnprintf(s_Return, s_BufferSize - 1, s_Format, s_VaArg) < s_BufferSize)
   {
    void *s_TempPtr = s_Return;
    s_Return = (void *)MZ_AllocString(s_TempPtr, (-1));
    MZ_Free(s_TempPtr);     
    break; 
   }
   else MZ_Free(s_Return); 
  }
  else break;
  s_Return = (void *)0;
  s_BufferSize += 256;
 }while(s_BufferSize < 8192); 
 return(s_Return);
#endif
}

void *MZ_SPrintf(const void *s_Format, ...)
{
 void *s_Return;
 va_list s_VaList;
 va_start(s_VaList, s_Format);
 s_Return = MZ_VSPrintf(s_Format, s_VaList);
 va_end(s_VaList);
 return(s_Return);
}

char *MZ_SearchString(const void *s_String, const void *s_This)
{
 int s_StringIndex, s_ThisIndex, s_EntryIndex, s_StringByte, s_ThisByte, s_ThisLength;
 if(s_String && s_This)
 {
  s_StringIndex = s_ThisIndex = s_EntryIndex = 0;
  s_ThisLength = MZ_StringLength(s_This);
  do
  {
   s_StringByte = MZ_PeekByte(s_String, s_StringIndex);
   s_ThisByte = MZ_PeekByte(s_This, s_ThisIndex);
   if(s_StringByte == '\0' || s_ThisByte == '\0')break;
   if(s_StringByte == s_ThisByte)
   {
    if(s_ThisIndex == 0)s_EntryIndex = s_StringIndex;
    if((++s_ThisIndex) == s_ThisLength)return((char *)__MZ_PEEK__(s_String, s_EntryIndex));
   }
   else s_ThisIndex = 0;
   s_StringIndex++;
  }while(1); 
 }
 return((char *)0);
}

char *MZ_StripString(const void *s_String, const void *s_StripString, int s_Flag)
{
 char *s_Return = (char *)0, *s_It;
 int s_Left, s_Right, s_StripLength;
 if(s_String && s_StripString)
 {
  s_Left = 0;
  s_Right = MZ_StringLength(s_String);
  s_StripLength = MZ_StringLength(s_StripString);
  if(s_Right > 0 && s_StripLength > 0)
  {
   if(s_Flag == 0)
   { /* All found */
    s_Return = MZ_AllocString(s_String, (-1));
    if(s_Return)
    {
     do
     {
      s_It = MZ_SearchString(__MZ_PEEK__(s_Return, s_Left), s_StripString);
      if(s_It)MZ_MoveMemory(s_It, __MZ_PEEK__(s_It, s_StripLength), MZ_StringLength(__MZ_PEEK__(s_It, s_StripLength)) + 1);
     }while(s_It);
#if 0 /* Strip memory = 1, Speed strip = 0 */
     s_It = s_Return;
     s_Return = MZ_AllocString(s_It, (-1));
     MZ_Free(s_It);
#endif
    }
   }
   else
   {
    if(s_Flag & 0x01)
    { /* Left */
     while(strncmp(__MZ_PEEK__(s_String, s_Left), s_StripString, s_StripLength) == 0)s_Left += s_StripLength;
    }
    if(s_Flag & 0x02)
    { /* Right */
     while(s_Left < (s_Right - s_StripLength))
     {
      if(strncmp(__MZ_PEEK__(s_String, s_Right - s_StripLength), s_StripString, s_StripLength) == 0)s_Right -= s_StripLength; 
      else break;
     }   
    }
    s_Return = MZ_AllocString(__MZ_PEEK__(s_String, s_Left), s_Right - s_Left);
   }
  }
  else s_Return = MZ_AllocString(s_String, (-1)); /* Drop string (New) */
 }
 return(s_Return);
}

#endif

/* End of source */
