/*
 Copyright (c) 2002 Information Equipment co.,LTD.
 All Right Reserved.

 Code by JaeHyuk Cho <minzkn@infoeq.co.kr> 

 Reference : man page's "statfs" 
*/

#if !defined(DEF_MZ_DiskSpace_c)
#define DEF_MZ_DiskSpace_c "MZ_DiskSpace.c"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <mntent.h>
#include <sys/time.h>
#include <sys/sysinfo.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/vfs.h>
#include <dirent.h>

#if !defined(t_u64)
#if defined(WIN32)
typedef double t_u64;
#else
typedef unsigned long long t_u64;
#endif
#endif

#define DEF_BUSY_CHECK (1)

#define DEF_UNIT_COMPUTE(x,u) \
         if(x > 99999lu) \
         { \
          x >>= 10; \
          s_Unit[(u)] = 'K'; \
          if(x > 99999lu) \
          { \
           x >>= 10; \
           s_Unit[(u)] = 'M'; \
           if(x > 99999lu) \
           { \
            x >>= 10; \
            s_Unit[(u)] = 'G'; \
            if(x > 99999lu) \
            { \
             x >>= 10; \
             s_Unit[(u)] = 'T'; \
	    } \
	   } \
	  } \
	 } \
         else s_Unit[(u)] = 'B'

int main(int s_Argc, char **s_Argv);
void *RealRootDevice(char *s_Name);
static void mstat(void);

int main(int s_Argc, char **s_Argv)
{
 int s_Return = (-1);
 struct sysinfo s_SysInfo;
 FILE *s_MountTable;
 fprintf(stdout, "MZ_DiskSpace v1.0.0 - Copyright(c)InfoEQ co.,LTD - %s %s\n", __DATE__, __TIME__);
 fprintf(stdout, "Code by JaeHyuk Cho - <minzkn@infoeq.co.kr>\n\n");
 if(s_Argc > 1)
 {
  if(strcasecmp(s_Argv[1], "mstat") == 0)
  {
   mstat(); 	  
   return(s_Return);
  }
 }
 
 fprintf(stdout, "System information\n");
 fprintf(stdout, "==================\n\n");
 if(sysinfo(&s_SysInfo) == 0)
 {
  fprintf(stdout, "%-16s = %ld %02ld:%02ld:%02ld\n", "Uptime",
          ((s_SysInfo.uptime / 60) / 60) / 24, 
          ((s_SysInfo.uptime / 60) / 60) % 24, 
          (s_SysInfo.uptime / 60)  % 60, 
          s_SysInfo.uptime % 60
	 );
  fprintf(stdout, "%-16s = %lu/%lu/%lu\n", "Loads 1/5/15min", s_SysInfo.loads[0], s_SysInfo.loads[1], s_SysInfo.loads[2]);
  fprintf(stdout, "%-16s = %lu\n", "TotalRam", s_SysInfo.totalram);
  fprintf(stdout, "%-16s = %lu\n", "FreeRam", s_SysInfo.freeram);
  fprintf(stdout, "%-16s = %lu\n", "SharedRam", s_SysInfo.sharedram);
  fprintf(stdout, "%-16s = %lu\n", "BufferRam", s_SysInfo.bufferram);
  fprintf(stdout, "%-16s = %lu\n", "TotalSwap", s_SysInfo.totalswap);
  fprintf(stdout, "%-16s = %lu\n", "FreeSwap", s_SysInfo.freeswap);
  fprintf(stdout, "%-16s = %u\n", "Process", s_SysInfo.procs);
 }
 s_MountTable = setmntent("/proc/mounts", "r");
 if(s_MountTable)
 {
  struct mntent *s_MountEntry;
  struct statfs s_StatusFS;
  fprintf(stdout, "\n");
  fprintf(stdout, "%-16s   %8s %6s %6s %6s %3s%% %s\n", "File system", "Type", "Total", "Used", "Avail", "Use", "Mounted on");
  fprintf(stdout, "----------------------------------------------------------------\n");
  while((s_MountEntry = getmntent(s_MountTable))) 
  {
   if(statfs(s_MountEntry->mnt_dir, &s_StatusFS) == 0)
   {
    if(s_StatusFS.f_blocks >= 0)
    {
     char *s_DeviceName;	     
     if(strcmp(s_MountEntry->mnt_fsname, "/dev/root") == 0)s_DeviceName = RealRootDevice("/");
     else s_DeviceName = strdup(s_MountEntry->mnt_fsname);
     if(s_DeviceName)
     {
      t_u64 s_TotalSize;
      t_u64 s_UseSize;
      t_u64 s_FreeSize;
      int s_Percent;
      char s_Unit[3];
      t_u64 s_Unit_TotalSize;
      t_u64 s_Unit_UseSize;
      t_u64 s_Unit_FreeSize;
      s_Unit_TotalSize = s_TotalSize = (t_u64)( (t_u64)s_StatusFS.f_blocks * (t_u64)s_StatusFS.f_bsize );      
      s_Unit_UseSize   = s_UseSize   = (t_u64)( (t_u64)(s_StatusFS.f_blocks - s_StatusFS.f_bfree) * (t_u64)s_StatusFS.f_bsize );
      s_Unit_FreeSize  = s_FreeSize  = (t_u64)( (t_u64)s_StatusFS.f_bavail * (t_u64)s_StatusFS.f_bsize );
      if(s_TotalSize > (t_u64)0)
      {
       s_Percent = (int)( (double)((double)(s_UseSize * (t_u64)100) / (double)(s_UseSize + s_FreeSize)) + 0.5 );
      }
      else s_Percent = 0;
      DEF_UNIT_COMPUTE(s_Unit_TotalSize, 0);
      DEF_UNIT_COMPUTE(s_Unit_UseSize, 1);
      DEF_UNIT_COMPUTE(s_Unit_FreeSize, 2);
#if defined(WIN32)
      fprintf(stdout, "%-16s %10s %5lu%c %5lu%c %5lu%c %3d%% %s\n",
              s_DeviceName,
	      s_MountEntry->mnt_type, /* MZ_GetFSName(s_StatusFS.f_type), */
              (unsigned long)s_Unit_TotalSize, s_Unit[0],
              (unsigned long)s_Unit_UseSize, s_Unit[1],
              (unsigned long)s_Unit_FreeSize, s_Unit[2],
	      s_Percent,
              s_MountEntry->mnt_dir
             );
#else
      fprintf(stdout, "%-16s %10s %5llu%c %5llu%c %5llu%c %3d%% %s\n",
              s_DeviceName,
	      s_MountEntry->mnt_type, /* MZ_GetFSName(s_StatusFS.f_type), */
              s_Unit_TotalSize, s_Unit[0],
              s_Unit_UseSize, s_Unit[1],
              s_Unit_FreeSize, s_Unit[2],
	      s_Percent,
              s_MountEntry->mnt_dir
             );
#endif
      free(s_DeviceName);
     }
    }
   }
  }
  fprintf(stdout, "----------------------------------------------------------------\n"
                  "Unit : B=1024^0, K=1024^1, M=1024^2, G=1024^3, T=1024^4\n"
	 );
  endmntent(s_MountTable);
 }
 fprintf(stdout, "\nEnd of DiskSpace\n");
 return(s_Return);
}

void *RealRootDevice(char *s_Name)
{
 void *s_Return = (void *)0;
 struct stat s_Status;
 struct stat s_StatusRoot;
 if(stat(s_Name, &s_StatusRoot) == 0)
 {
  dev_t s_Device;
  DIR *s_Dir;
  char *s_DeviceDirectory = "/dev";
  s_Device = s_StatusRoot.st_rdev;
  if(s_Device == 0)s_Device = s_StatusRoot.st_dev;
  s_Dir = opendir(s_DeviceDirectory);
  if(s_Dir)
  {
   struct dirent *s_DirectoryEntry;
   do
   {
    s_DirectoryEntry = readdir(s_Dir);
    if(s_DirectoryEntry)
    {
     if(strcmp(s_DirectoryEntry->d_name, "..") == 0)continue;
     s_Return = malloc(strlen(s_DeviceDirectory) + 1 + strlen(s_DirectoryEntry->d_name) + 1);
     if(s_Return)
     {
      sprintf(s_Return, "%s/%s", s_DeviceDirectory, s_DirectoryEntry->d_name);
      if(stat(s_Return, &s_Status) == 0)
      {
       if(S_ISBLK(s_Status.st_mode) != 0)
       {
        if(s_Status.st_rdev == s_Device)break;
       }
      }
      free(s_Return);
      s_Return = (void *)0;      
     }
    }
   }while(s_DirectoryEntry); 
   closedir(s_Dir);
  }
  else fprintf(stderr, "Can not open directory \"%s\"\n", s_DeviceDirectory);
 }
 else fprintf(stderr, "Can not stat \"/\"\n");
 if(s_Return == (void *)0)s_Return = strdup("/dev/root");
 return(s_Return);
}

static int mstat_check(long s_Old, long s_New, long s_M)
{
 int s_Return = 0;
 long s_Min, s_Max;
 s_Min = s_Old - s_M;
 s_Max = s_Old + s_M;
 if(s_Min < s_New && s_Max > s_New)s_Return = 0; 
 else s_Return = 1;
 return(s_Return);
}

#define DEF_MEMCHECK_UNIT (4 << 10)
#define DEF_MEMCHECK(O,N,M) mstat_check((O), (N), (M)) 
// (! (((O) + (M)) > N && ((O) - (((O) >= (M)) ? (M) : 0)) < N) )

static void mstat(void)
{
 struct sysinfo s_SysInfo;
 struct sysinfo s_OldInfo;
 int s_SyncTime = 0;
 memset(&s_OldInfo, 0, sizeof(struct sysinfo));
 do
 {
  if(sysinfo(&s_SysInfo) == 0)
  {
#if DEF_BUSY_CHECK == (1)
   if(DEF_MEMCHECK(s_OldInfo.totalram, s_SysInfo.totalram, DEF_MEMCHECK_UNIT) || 	  
      DEF_MEMCHECK(s_OldInfo.freeram, s_SysInfo.freeram, DEF_MEMCHECK_UNIT) || 	  
      DEF_MEMCHECK(s_OldInfo.sharedram, s_SysInfo.sharedram, DEF_MEMCHECK_UNIT) || 	  
      DEF_MEMCHECK(s_OldInfo.bufferram, s_SysInfo.bufferram, DEF_MEMCHECK_UNIT) || 	  
      DEF_MEMCHECK(s_OldInfo.totalram, s_SysInfo.totalram, DEF_MEMCHECK_UNIT) || 	  
      DEF_MEMCHECK(s_OldInfo.freeram, s_SysInfo.freeram, DEF_MEMCHECK_UNIT) || 	  
      s_OldInfo.procs     != s_SysInfo.procs  
     )
#else
   if(s_OldInfo.totalram  != s_SysInfo.totalram   || 
      s_OldInfo.freeram   != s_SysInfo.freeram    || 
      s_OldInfo.sharedram != s_SysInfo.sharedram  || 
      s_OldInfo.bufferram != s_SysInfo.bufferram  || 
      s_OldInfo.totalswap != s_SysInfo.totalswap  || 
      s_OldInfo.freeswap  != s_SysInfo.freeswap   || 
      s_OldInfo.procs     != s_SysInfo.procs  
     )
#endif
   {	   
    if(s_SyncTime == 0)
    {
     fprintf(stdout, 
         "%8s %6s %10s %10s %10s %10s %10s %10s %6s\n",
	 "Uptime",
	 "Loads1",
	 "TotalRam",
	 "FreeRam",
	 "SharedRam",
	 "BufferRam",
	 "TotalSwap",
	 "FreeSwap",
	 "Procs"
	);
    }
    fprintf(stdout,
            "%8lu %6lu %10lu %10lu %10lu %10lu %10lu %10lu %6d\n",
	    s_SysInfo.uptime,
	    s_SysInfo.loads[0],
	    s_SysInfo.totalram,
	    s_SysInfo.freeram,
	    s_SysInfo.sharedram,
	    s_SysInfo.bufferram,
	    s_SysInfo.totalswap,
	    s_SysInfo.freeswap,
	    s_SysInfo.procs 
	   );
    fflush(stdout);
    s_SyncTime++;
    if(s_SyncTime > 5)s_SyncTime = 0;
    memcpy(&s_OldInfo, &s_SysInfo, sizeof(struct sysinfo));	   
   }
  }
  else fprintf(stdout, "Can not sysinfo !!!\n");
#if DEF_BUSY_CHECK == (1)
  usleep(10000);
#else
  sleep(1);
#endif
 }while(1);
}

#endif 
	

/* End of source */ 
