/*
 Copyright (c) 2002 Information Equipment co.,LTD.
 All Right Reserved.

 Code by JaeHyuk Cho <minzkn@infoeq.co.kr> 

 Graphics Device Control library v2.0.0-demo1 [ Optimized for fire code ]


 This example !!!
   - Only 16bit frame buffer support.
   - Simple is best.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#define DEF_USE_OPTIMIZE_FIRE (1)

#define DEF_FBIOGET_VSCREENINFO         0x4600
#define DEF_FBIOGET_FSCREENINFO         0x4602

typedef struct ts_DFB
{
 int                                  Handle;
 int                                  ResX, ResY, ResB, RealX, RealY;
 unsigned long                        MapSize;
 void                                *DirectMap;
 int                                 *Table;
}t_DFB;

typedef struct
{
 char                               ID[16];                        /* Identification */
 void                              *PhysicalAddress;               /* Start of physical memory address */
 unsigned long                      PhysicalLength;                /* Length of frame buffer memory */
 unsigned long                      Type;                          /* Frame buffer type */  
 unsigned long                      TypeAUX;                       /* Interleave for interleaved planes */
 unsigned long                      TypeVisual;                    /* Visual type */
 unsigned short                     StepX;                         /* Step X */
 unsigned short                     StepY;                         /* Step Y */
 unsigned short                     StepXY;                        /* Step X & Y */
 unsigned long                      LineLength;                    /* Length of line in bytes */
 void                              *IOAddress;                     /* Start of I/O map address */
 unsigned long                      IOLength;                      /* I/O mapping limited */
 unsigned long                      Accel;                         /* Accel type */
 unsigned short                     Reserved[3];                   /* Reserved information area */
}t_DFB_INFO_FB_FIX;

typedef struct
{
 unsigned long                      Offset, Length, Right;         /* Most significant bit is right */ 
}t_DFB_FB_BIT;
 
typedef struct 
{
 unsigned long                      ResX, ResY;
 unsigned long                      VirtualX, VirtualY;
 unsigned long                      OffsetX, OffsetY;
 unsigned long                      BPP;                           /* Bits per pixel */
 unsigned long                      GrayScale;
 t_DFB_FB_BIT                       RGB[3];
 t_DFB_FB_BIT                       Transparentcy;
 unsigned long                      IsStandard;
 unsigned long                      Activate;
 unsigned long                      Height, Width;
 unsigned long                      AccelFlags;
 unsigned long                      PixClock;
 unsigned long                      LeftMargin, RightMargin, TopMargin, BottomMargin;
 unsigned long                      HSyncLength, VSyncLength, Sync;
 unsigned long                      VMode;
 unsigned long                      Reserved[6];  
}t_DFB_INFO_FB_VAR;

#if DEF_USE_OPTIMIZE_FIRE == (1)
static t_DFB *MZ_DFB_Close(t_DFB *s_Handle)
{
 if(s_Handle)
 {
  if(s_Handle->Table)free(s_Handle->Table);
  if(s_Handle->DirectMap && s_Handle->MapSize > 0lu)munmap(s_Handle->DirectMap, s_Handle->MapSize);
  if(s_Handle->Handle >= 0)close(s_Handle->Handle);
  free(s_Handle);
 }
 /* else Error!!! */
 return(s_Handle);
}
#endif

static int MZ_DFB_Detect(t_DFB *s_Handle)
{
 int s_Return = (-1);
 if(s_Handle)
 {
  if(s_Handle->Handle >= 0)
  {
   t_DFB_INFO_FB_FIX s_Info_FIX;	  
   t_DFB_INFO_FB_VAR s_Info_VAR;	  
   if(ioctl(s_Handle->Handle, DEF_FBIOGET_FSCREENINFO, &s_Info_FIX) == 0)
   {
    if(ioctl(s_Handle->Handle, DEF_FBIOGET_VSCREENINFO, &s_Info_VAR) == 0)
    {
     unsigned long s_AvailMapSize;    
     s_Handle->RealX = (int)s_Info_VAR.ResX;
     s_Handle->ResX = (int)(s_Info_FIX.LineLength / (s_Info_VAR.BPP >> 3));
     s_Handle->RealY = s_Handle->ResY = (int)s_Info_VAR.ResY;
     s_Handle->ResB = (int)s_Info_VAR.BPP >> 3;
     s_Handle->MapSize = (unsigned long)s_Handle->ResX * (unsigned long)s_Handle->ResY * (unsigned long)s_Handle->ResB;
     s_AvailMapSize    = (unsigned long)s_Handle->ResX * (unsigned long)s_Handle->ResY * (unsigned long)(s_Info_VAR.BPP & 0x07lu);
     s_Handle->MapSize+= (s_AvailMapSize >> 3) + ((s_AvailMapSize & 0x07) ? 1 : 0);
     s_Handle->DirectMap = mmap((void *)0, s_Handle->MapSize, PROT_READ | PROT_WRITE, MAP_SHARED, s_Handle->Handle, 0); 
     if(s_Handle->DirectMap != (void *)(-1))
     {
      s_Handle->DirectMap = s_Handle->DirectMap;
      if(s_Handle->Table)free(s_Handle->Table);
      s_Handle->Table = (int *)malloc(s_Handle->ResY * sizeof(int));
      if(s_Handle->Table)
      {
       for(s_Return = 0;s_Return < s_Handle->ResY;s_Return++)*(s_Handle->Table + s_Return) = s_Return * s_Handle->ResX * s_Handle->ResB; 
       s_Return = 1;
      }
      /* else Error!!! */
     }
     /* else Error!!! */
    }
    /* else Error!!! */
   }
   /* else Error!!! */
  }
  /* else Error!!! */
 }
 /* else Error!!! */
 return(s_Return);
}

static t_DFB *MZ_DFB_Open(void)
{
 t_DFB *s_Return = (t_DFB *)malloc(sizeof(t_DFB));
 if(s_Return)
 {
  t_DFB *s_Handle = s_Return;	 
  /* memset((void *)s_Return, 0, sizeof(t_DFB)); */
  s_Handle->Handle = open("/dev/fb0", O_RDWR);
  if(s_Handle->Handle >= 0)
  {   
   if(MZ_DFB_Detect(s_Handle) >= 0)s_Return = s_Handle;
   /* else Error!!! */
  }
  /* else Error!!! */
 }
 /* else Error!!! */
 return(s_Return);
}

int main(void)
{
 t_DFB *s_DFB = MZ_DFB_Open();
 if(s_DFB)
 {
  unsigned char *s_ColorMAP = (unsigned char *)malloc(s_DFB->ResX * s_DFB->ResY);
#if DEF_USE_OPTIMIZE_FIRE == (0)
  if(s_DFB->ResB != 2)
  {
   fprintf(stdout, "Not support %d bit color !!!\n", s_DFB->ResB << 3);
   if(s_ColorMAP)free(s_ColorMAP);
   MZ_DFB_Close(s_DFB);
   exit(0);
  }
#endif
  if(s_ColorMAP)
  {
   int s_Entry, s_Limit, s_FrameCount = 0, s_ColorType = 0, s_ColorTable[256], s_Count;
   /* memset(s_ColorMAP, 0, s_DFB->ResX * s_DFB->ResY); */
   s_Limit = s_DFB->ResX * (s_DFB->ResY - 128);
   for(s_Count = 0;s_Count < 64;s_Count++)s_ColorTable[s_Count + 192] = ((63 >> 1) << 11) | (63 << 5) | (63 >> 1);
   do
   {
    if(s_FrameCount == 0)
    {
     if(s_ColorType == 0)
     {	     
      for(s_Count = 0;s_Count < 64;s_Count++)s_ColorTable[s_Count +   0] = (s_Count >> 1) << 11;
      for(s_Count = 0;s_Count < 64;s_Count++)s_ColorTable[s_Count +  64] = ((63 >> 1) << 11) | (s_Count << 5);
      for(s_Count = 0;s_Count < 64;s_Count++)s_ColorTable[s_Count + 128] = ((63 >> 1) << 11) | (63 << 5) | (s_Count >> 1);
     }
     else if(s_ColorType == 1)
     {
      for(s_Count = 0;s_Count < 64;s_Count++)s_ColorTable[s_Count +   0] = s_Count << 5;
      for(s_Count = 0;s_Count < 64;s_Count++)s_ColorTable[s_Count +  64] = (63 << 5) | (s_Count >> 1);
      for(s_Count = 0;s_Count < 64;s_Count++)s_ColorTable[s_Count + 128] = ((s_Count >> 1) << 11) | (63 << 5) | (63 >> 1);
     }
     else
     {
      for(s_Count = 0;s_Count < 64;s_Count++)s_ColorTable[s_Count +   0] = (s_Count >> 1);
      for(s_Count = 0;s_Count < 64;s_Count++)s_ColorTable[s_Count +  64] = (s_Count << 5) | (63 >> 1);
      for(s_Count = 0;s_Count < 64;s_Count++)s_ColorTable[s_Count + 128] = ((s_Count >> 1) << 11) | (63 << 5) | (63 >> 1);
     }
    }
    s_Entry = (s_DFB->ResX * (s_DFB->ResY - 1));   
    for(s_Count = 0;s_Count < (s_DFB->RealX >> 3);s_Count++)s_ColorMAP[s_Entry + (rand() % s_DFB->RealX)] = s_ColorTable[(rand() & 0xbf) + 10];
    s_Count = (s_DFB->ResX * (s_DFB->ResY - 2)) + s_DFB->ResX - 1; 
    s_Entry = s_Count + s_DFB->ResX - 1;	
    do
    {
      s_ColorMAP[s_Count] = ((s_ColorMAP[s_Count    ] + 
	                     s_ColorMAP[s_Entry    ] + 
	                     s_ColorMAP[s_Entry + 1] + 
                             s_ColorMAP[s_Entry + 2]) >> 2); 
     if(s_ColorMAP[s_Count] > 0)s_ColorMAP[s_Count]--;
     *(((unsigned short *)s_DFB->DirectMap) + s_Count) = (unsigned short)s_ColorTable[s_ColorMAP[s_Count]];
     if(s_Count < s_Limit)break;
     s_Entry--;	    
    }while(--s_Count > 0);
    s_FrameCount++;	  
    if(s_FrameCount > 128)
    {
     s_FrameCount = 0, s_ColorType++;
     if(s_ColorType >= 3)s_ColorType = 0;
    } 
    usleep(10000);
   }while(1);
#if DEF_USE_OPTIMIZE_FIRE == (1)
   free(s_ColorMAP);
#endif
  }
#if DEF_USE_OPTIMIZE_FIRE == (1)
  s_DFB = MZ_DFB_Close(s_DFB);  
#endif  
 }
 return(0);
}

/* End of source */ 
