/*
 Copyright (c) 2002 Information Equipment co.,LTD.
 All Right Reserved.

 Code by JaeHyuk Cho <minzkn@infoeq.com> 
*/

#if !defined(DEF_mixer_c)
#define DEF_mixer_c "mixer.c"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>

#include "mixer.h"

#define DEF_AC97_MIXER_TABLE(x) {E_AC97_MIXER_##x, E_AC97_MIXER_GET_##x, E_AC97_MIXER_SET_##x}
#define DEF_AC97_MIXER_INDEX(x) E_AC97_MIXER_##x 

int MZ_Mixer_AC97(char *s_DeviceName, unsigned long s_Command, unsigned long s_Type, int *s_Left, int *s_Right);
int MZ_Index_AC97(unsigned long s_Type);
char *MZ_Name_AC97(unsigned long s_Type);

int g_AC97_Index[] = {
     DEF_AC97_MIXER_INDEX(VOLUME), DEF_AC97_MIXER_INDEX(BASS), DEF_AC97_MIXER_INDEX(TREBLE),
     DEF_AC97_MIXER_INDEX(SYNTH), DEF_AC97_MIXER_INDEX(PCM), DEF_AC97_MIXER_INDEX(SPEAKER),
     DEF_AC97_MIXER_INDEX(LINE), DEF_AC97_MIXER_INDEX(MIC), DEF_AC97_MIXER_INDEX(CD),
     DEF_AC97_MIXER_INDEX(IMIX), DEF_AC97_MIXER_INDEX(ALTPCM), DEF_AC97_MIXER_INDEX(RECLEV),
     DEF_AC97_MIXER_INDEX(IGAIN), DEF_AC97_MIXER_INDEX(OGAIN), DEF_AC97_MIXER_INDEX(LINE1),
     DEF_AC97_MIXER_INDEX(LINE2), DEF_AC97_MIXER_INDEX(LINE3), DEF_AC97_MIXER_INDEX(DIGITAL1),
     DEF_AC97_MIXER_INDEX(DIGITAL2), DEF_AC97_MIXER_INDEX(DIGITAL3), DEF_AC97_MIXER_INDEX(PHONEIN),
     DEF_AC97_MIXER_INDEX(PHONEOUT), DEF_AC97_MIXER_INDEX(VIDEO), DEF_AC97_MIXER_INDEX(RADIO),
     DEF_AC97_MIXER_INDEX(MONITOR), DEF_AC97_MIXER_INDEX(GPIO), DEF_AC97_MIXER_INDEX(REC),
     DEF_AC97_MIXER_INDEX(ONOFF_MIN), DEF_AC97_MIXER_INDEX(ONOFF_MAX), DEF_AC97_MIXER_INDEX(NONE)
    };	  

char *g_AC97_Name[] = {
       "master", "bass", "treble", "synth", "pcm", "speaker", "line", "mic", "cd", "imix",
       "altpcm", "reclev", "igain", "ogain", "line1", "line2", "line3", "digital1", "digital2",
       "digital3", "phonein", "phoneout", "video", "radio", "monitor", "gpio", "rec", "onoff_min",
       "onoff_max", "none", (char *)0
      };

int MZ_Mixer_AC97(char *s_DeviceName, unsigned long s_Command, unsigned long s_Type, int *s_Left, int *s_Right)
{
 int s_Return = (-1);
 int s_Handle;
 if(s_DeviceName == (char *)0)s_DeviceName = "/dev/mixer";
 s_Handle = open(s_DeviceName, O_RDONLY);
 if(s_Handle >= 0)
 {
  const t_Mixer_AC97_Table s_MixerTable[] = {
                      DEF_AC97_MIXER_TABLE(VOLUME), DEF_AC97_MIXER_TABLE(BASS), DEF_AC97_MIXER_TABLE(TREBLE),
                      DEF_AC97_MIXER_TABLE(SYNTH), DEF_AC97_MIXER_TABLE(PCM), DEF_AC97_MIXER_TABLE(SPEAKER),
                      DEF_AC97_MIXER_TABLE(LINE), DEF_AC97_MIXER_TABLE(MIC), DEF_AC97_MIXER_TABLE(CD),
                      DEF_AC97_MIXER_TABLE(IMIX), DEF_AC97_MIXER_TABLE(ALTPCM), DEF_AC97_MIXER_TABLE(RECLEV),
                      DEF_AC97_MIXER_TABLE(IGAIN), DEF_AC97_MIXER_TABLE(OGAIN), DEF_AC97_MIXER_TABLE(LINE1),
                      DEF_AC97_MIXER_TABLE(LINE2), DEF_AC97_MIXER_TABLE(LINE3), DEF_AC97_MIXER_TABLE(DIGITAL1),
                      DEF_AC97_MIXER_TABLE(DIGITAL2), DEF_AC97_MIXER_TABLE(DIGITAL3), DEF_AC97_MIXER_TABLE(PHONEIN),
                      DEF_AC97_MIXER_TABLE(PHONEOUT), DEF_AC97_MIXER_TABLE(VIDEO), DEF_AC97_MIXER_TABLE(RADIO),
                      DEF_AC97_MIXER_TABLE(MONITOR), DEF_AC97_MIXER_TABLE(GPIO), DEF_AC97_MIXER_TABLE(REC),
                      DEF_AC97_MIXER_TABLE(ONOFF_MIN), DEF_AC97_MIXER_TABLE(ONOFF_MAX), DEF_AC97_MIXER_TABLE(NONE)
                     };	  
  int s_Index = 0;
  int s_Level;
  while(s_MixerTable[s_Index].Type != E_AC97_MIXER_NONE)
  {
   if((s_Type & s_MixerTable[s_Index].Type) == s_MixerTable[s_Index].Type)
   {
    if(s_Command & E_AC97_VOLUME || s_Command & E_AC97_VOLUME_ADJUST)
    {
     s_Level = 0;
     if(s_Left == (int *)0 || s_Right == (int *)0 || s_Command & E_AC97_VOLUME_ADJUST)
     {
      if(ioctl(s_Handle, s_MixerTable[s_Index].GetCtrl, &s_Level) == 0)
      {
       s_Level &= 0x7fff;	      
      }
      else fprintf(stderr, "mixer.c: MZ_Mixer_AC97 - [ERROR] Can not get pre-volume !!!\n");
     }
     if(s_Command & E_AC97_VOLUME_ADJUST)
     {
      int s_Volume[2];
      if(s_Left)s_Volume[0] = *(s_Left);
      else s_Volume[0] = 0;
      if(s_Right)s_Volume[1] = *(s_Right);
      else s_Volume[1] = 0;
      s_Volume[0] = (s_Level & 0xff) + s_Volume[0]; 
      if(s_Volume[0] > 100)s_Volume[0] = 100;
      if(s_Volume[0] < 0)s_Volume[0] = 0;
      s_Volume[1] = ((s_Level >> 8) & 0xff) + s_Volume[1];
      if(s_Volume[1] > 100)s_Volume[1] = 100;
      if(s_Volume[1] < 0)s_Volume[1] = 0;
      s_Level = s_Volume[0] | (s_Volume[1] << 8); 
     }
     else
     {
      if(s_Left)s_Level |= (int)((*(s_Left)) & 0xff);
      if(s_Right)s_Level |= (int)((*(s_Right) << 8) & 0xff00);
     }
     if(ioctl(s_Handle, s_MixerTable[s_Index].SetCtrl, &s_Level) == 0)
     {
      
     }
     else fprintf(stderr, "mixer.c: MZ_Mixer_AC97 - [ERROR] Can not set volume !!!\n");
    }
    if(s_Command & E_AC97_MUTE_ON || s_Command & E_AC97_MUTE_OFF)
    {
     if(ioctl(s_Handle, s_MixerTable[s_Index].GetCtrl, &s_Level) == 0)
     {
      if(s_Command & E_AC97_MUTE_ON)s_Level |= 0x8000; /* Mute bit */
      if(s_Command & E_AC97_MUTE_OFF)s_Level &= 0x7fff; /* Mute mask */
      if(ioctl(s_Handle, s_MixerTable[s_Index].SetCtrl, &s_Level) == 0)
      {
      
      }
      else fprintf(stderr, "mixer.c: MZ_Mixer_AC97 - [ERROR] Can not set mute (Mute) !!!\n");
     }
     else fprintf(stderr, "mixer.c: MZ_Mixer_AC97 - [ERROR] Can not get mute (Mute) !!!\n");
    }
    if(s_Command & E_AC97_GETVOLUME || s_Command & E_AC97_GETMUTE)
    {
     if(ioctl(s_Handle, s_MixerTable[s_Index].GetCtrl, &s_Level) == 0)
     {
      if(s_Command & E_AC97_GETVOLUME)s_Level &= 0x7fff;	       
      if(s_Left)*(s_Left) = (int)(s_Level & 0xff);
      if(s_Right)*(s_Right) = (int)((s_Level >> 8) & 0xff);
     }
     else 
     {
      if(s_Left)*(s_Left) = (-1);
      if(s_Right)*(s_Right) = (-1);
     }
    }
   }
   s_Index++;	   
  }
  close(s_Handle);
 }
 else fprintf(stderr, "mixer.c: MZ_Mixer_AC97 - [ERROR] Can not open mixer !!!\n");
 return(s_Return);
}

int MZ_Index_AC97(unsigned long s_Type)
{
 int s_Return = 0;
 while(g_AC97_Index[s_Return] != E_AC97_MIXER_NONE)
 {
  if(g_AC97_Index[s_Return] == s_Type)break; 	 
  s_Return++;
 }
 if(g_AC97_Index[s_Return] == E_AC97_MIXER_NONE)s_Return = (-1);
 return(s_Return);
}

char *MZ_Name_AC97(unsigned long s_Type)
{
 char *s_Return = "Unknown";
 int s_Index = MZ_Index_AC97(s_Type);
 if(s_Index >= 0)s_Return = g_AC97_Name[s_Index];
 return(s_Return);
}

#endif 

/* End of source */ 
